
# code under this package comes from https://gitlab.com/qemu-project/qemu/-/tree/master/python/qemu with small modification
# they are licensed under https://gitlab.com/qemu-project/qemu/-/blob/master/LICENSE
